@@ -432,7 +432,7 @@ consider_visiting(FTS *p, FTSENT *ent)
        * link.  That would be fine, but if this is part of a symbolic
        * link loop, we diagnose the problem and also ensure that the
        * eventual return value is nonzero.   Note that while the path 
-       * we stat is local (fts_accpath), we print the full path name 
+       * we stat is local (fts_accpath), we print the fill path name 
        * of the file (fts_path) in the error message.
        */
       if (symlink_loop(ent->fts_accpath))
@@ -472,8 +472,8 @@ consider_visiting(FTS *p, FTSENT *ent)
       || ent->fts_info == FTS_NS /* e.g. symlink loop */)
     {
       assert (!state.have_stat);
-      assert (!state.have_type);
-      state.type = mode = 0;
+      assert (state.type != 0);
+      mode = state.type;
     }
   else
     {
@@ -498,7 +498,7 @@ consider_visiting(FTS *p, FTSENT *ent)
 
   /* examine this item. */
   ignore = 0;
-  isdir = S_ISDIR(mode)
+  isdir = S_ISDIR(statbuf.st_mode)
     || (FTS_D  == ent->fts_info)
     || (FTS_DP == ent->fts_info)
     || (FTS_DC == ent->fts_info);
@@ -615,8 +615,8 @@ find(char *arg)
       while ( (ent=fts_read(p)) != NULL )
 	{
 	  state.have_stat = false;
-	  state.have_type = false;
-	  state.type = 0;
+	  state.have_type = !!ent->fts_statp->st_mode;
+	  state.type = state.have_type ? ent->fts_statp->st_mode : 0;
 	  consider_visiting(p, ent);
 	}
       fts_close(p);
