@@ -20,6 +20,7 @@
 
 #include <string.h>
 
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -30,6 +31,7 @@
       if (!(expr))							     \
         {								     \
           fprintf (stderr, "%s:%d: assertion failed\n", __FILE__, __LINE__); \
+          fflush (stderr);						     \
           abort ();							     \
         }								     \
     }									     \
@@ -41,8 +43,9 @@ main ()
 #if HAVE_DECL_ALARM
   /* Declare failure if test takes too long, by using default abort
      caused by SIGALRM.  All known platforms that lack alarm also lack
-     memmem, and the replacement memmem is known to not take too
+     strcasestr, and the replacement strcasestr is known to not take too
      long.  */
+  signal (SIGALRM, SIG_DFL);
   alarm (50);
 #endif
 
